/**
 * Functions that interact with the database
 *
 * @author Henrique Salvadori Coelho
 *         henriquesc@gmail.com
 *         http://hcoelho.com
 *         https://github.com/hscasn
 *
 * **************************************************
 *
 * @typedef {{
 *   table_name: string,
 *   column_name: string,
 *   referenced_table_name: string,
 *   referenced_column_name: string
 * }} Constraint
 */

const PG = require('pg');

const tableValidation = require('./tableValidation');
const isValidTable = tableValidation.isValidTable;
const getCreationQuery = tableValidation.getCreationQuery;
const getAlterQuery = tableValidation.getAlterQuery;
const getForeignKeyCreationString = tableValidation.getForeignKeyCreationString;
const getIndexCreationString = tableValidation.getIndexCreationString;

let settings = undefined;
let log = () => {};
let verbose = false;


/**
 * Creates the connection with the database and also creates the handler to disconnected
 */
function connect() {
  const pool = new PG.Pool({
    host: settings.connectionOptions.host,
    user: settings.connectionOptions.user,
    password: settings.connectionOptions.password,
    database: settings.connectionOptions.database,
    port: settings.connectionOptions.port,
  });

  settings.connection = pool;
  pool.on('error', (e) => {
    log(e);
    if (!settings.connection) {
      throw 'The connection with the database could not be completed. The host and the credentials may be incorrect.';
    }
  });
}


/**
 * When called, will end the connection with the database
 */
function disconnect() {
  try { settings.connection.end(); }
  catch (e) {}
  verbose && log('Connection terminated');
}


/**
 * Navigates through all the tables described in the structure and expands the ones in the database
 *
 * @param {Function} _log
 * @param {Settings} _settings
 * @param {string} _settings.connectionOptions.host Connection username
 * @param {string} _settings.connectionOptions.user Connection username
 * @param {string} _settings.connectionOptions.password Connection password
 * @param {string} _settings.connectionOptions.database Name of the database
 * @param {number} _settings.connectionOptions.port Port for the connection
 * @param {number} _settings.connectionOptions.connectTimeout Milliseconds until a timeout occurs
 *
 * Creates:
 * @param {Connection} _settings.connection Will be used to make the queries
 * @param {function} _settings.disconnect When called, will terminate the connection
 */
async function expand(_settings, _log, _verbose) {
  settings = _settings;
  log = _log;
  verbose = _verbose;

  if (settings.structure.constructor !== Array) {
    throw 'The structure of the database must be an array'
  }

  if (settings.structure.length <= 0) { return; }

  connect();

  try {
    for (let i = 0; i < settings.structure.length; i++) {
      await searchTable(settings.structure[i]);
    }
    await searchForeignKeys();
    await searchIndexes();
  }
  catch (e) {
    log(e);
  }
  finally {
    await disconnect();
  }
}


/**
 * Looks for a table in the database and creates or expands it, depending if the table exists or not
 *
 * @param {Table} table is one of the entries from the structures
 * @param {string} table.name is the name of the table
 * @returns {Promise}
 */
async function searchTable(table) {

  if (!isValidTable(table)) { return undefined; }

  try {
    verbose && log(`Searching for table ${table.name}`);
    const query = `SELECT * FROM pg_catalog.pg_tables WHERE tablename='${table.name}'`;
    const {rows} = await settings.connection.query(query);

    if (rows.length > 0) {
      verbose && log(`Table ${table.name} exists`);
      await searchColumns(table);
    }
    else {
      verbose && log(`Table ${table.name} does not exist`);
      await createTable(table);
    }

  }
  catch (e) {
    log(e);
  }

}


/**
 * Creates a table in the database
 *
 * @param {Table} table
 * @returns {Promise}
 */
async function createTable(table) {
  try {
    const query = getCreationQuery(table);
    await settings.connection.query(query);
    log(`Table ${table.name} created`);
  }
  catch (e) {
    log(`Error while creating the table ${table.name}. ` +
      `The structure for this table may be incorrect`);
    log(e.message);
  }
}


/**
 * Searches within the table for columns that are not present
 *
 * @param {Table} table
 * @returns {Promise}
 */
async function searchColumns(table) {
  try {
    const {rows} = await settings.connection.query(`SELECT column_name FROM INFORMATION_SCHEMA.COLUMNS where table_name='${table.name}';`);
    const columnsInTable = rows.map(column => column.column_name.toLowerCase());

    if (!table.columns || table.columns.constructor !== Array) {
      return;
    }

    const columnsToCreate = [];

    // Finding which columns do not exist
    table.columns.forEach(column => {
      if (columnsInTable.includes(column.name.toLowerCase())) {
        verbose && log(`Column ${column.name} in table ${table.name} already exists`);
      }
      else {
        verbose && log(`Column ${column.name} in table ${table.name} does not exist`);
        columnsToCreate.push(column);
      }
    });

    await createColumns(table, columnsToCreate);
  }
  catch (e) {
    log(e);
  }
}


/**
 * Adds the specified columns in a table
 *
 * @param {Table} table
 * @param {[Column]} columns
 * @returns {Promise}
 */
async function createColumns(table, columns) {
  if (!columns || columns.length < 1) { return undefined; }

  const sql = getAlterQuery(table, columns);
  try {
    await settings.connection.query(sql);
    log(`Table ${table.name} altered. Columns added: ${table.columns.map(c => c.name).join(', ')}`);
  }
  catch (e) {
    log(`Error while adding columns for ${table.name}. ` +
      `The structure for this table may be incorrect`);
    log(e.message);
    log(sql);
  }

}


async function searchForeignKeys() {

  // Query to get all the foreign keys
  const query = (
    `SELECT ` +
    `    tc.table_name AS TABLE_NAME, ` +
    `    kcu.column_name AS COLUMN_NAME,  ` +
    `    ccu.table_name AS REFERENCED_TABLE_NAME, ` +
    `    ccu.column_name AS REFERENCED_COLUMN_NAME ` +
    `FROM  ` +
    `    information_schema.table_constraints AS tc  ` +
    `    JOIN information_schema.key_column_usage AS kcu ` +
    `      ON tc.constraint_name = kcu.constraint_name ` +
    `    JOIN information_schema.constraint_column_usage AS ccu ` +
    `      ON ccu.constraint_name = tc.constraint_name ` +
    `WHERE constraint_type = 'FOREIGN KEY';`
  );

  try {
    const results = await settings.connection.query(query);
    const constraintList = parseResults(results.rows);
    const allConstraints = getStructureConstraints(constraintList);
    const fks = getFKsToAdd(allConstraints);
    const queries = generateQueries(fks);
    await performQueries(queries);
  }
  catch (e) {
    log(e);
  }

  /**
   * Receives the results from the database and parses them into a more friendly object
   * (object of Table.name:[Foreign Keys])
   *
   * @param {[]} rows
   * @returns {{string: [ForeignKey]}}
   */
  function parseResults(rows) {
    verbose && log(`Checking foreign keys`);

    const constraintList = {};

    rows.forEach(constraint => {
      if (!constraintList[constraint.table_name]) { constraintList[constraint.table_name] = []; }
      constraintList[constraint.table_name].push({
        column: constraint.column_name,
        referenceTable: constraint.referenced_table_name,
        referenceColumn: constraint.referenced_column_name
      });
    });

    return constraintList;
  }


  /**
   * Merges the parsed results from the database with the constraints specified in the structure, using the same
   * object structure
   *
   * @param constraintsInDb
   * @returns {{ inDb: {string: [ForeignKey]}, inStructure: {string: [ForeignKey]} }}
   */
  function getStructureConstraints(constraintsInDb) {
    const allConstraints = {};

    allConstraints.inDb = constraintsInDb;
    allConstraints.inStructure = {};

    settings.structure.forEach(table => {
      if (table.foreignKeys && table.foreignKeys.constructor === Array) {
        allConstraints.inStructure[table.name] = table.foreignKeys;
      }
    });

    return allConstraints;
  }


  /**
   * Receives the merged results from getStructureConstraints, compares them, and isolates only the ones that are not
   * present in the DB and need to be added
   *
   * @param {{ inDb: {string: [ForeignKey]}, inStructure: {string: [ForeignKey]} }} constraints
   * @returns {{string: [ForeignKey]}}
   */
  function getFKsToAdd(constraints) {
    const toAdd = {};

    Object.keys(constraints.inStructure).forEach(key => {
      const inDb = constraints.inDb[key];
      const inStructure = constraints.inStructure[key];

      const filtered = inStructure.filter(constraint => !contains(inDb, constraint));

      if (filtered.length > 0) { toAdd[key] = filtered; }
    });

    return toAdd;

    /**
     *
     * @param {[ForeignKey]} haystack
     * @param {ForeignKey} needle
     * @return {boolean}
     */
    function contains(haystack, needle) {
      if (!haystack) { return false; }
      for (let i = 0; i < haystack.length; i++) {
        const hay = haystack[i];
        if (hay.column === needle.column &&
          hay.referenceColumn === needle.referenceColumn &&
          hay.referenceTable === needle.referenceTable) { return true; }
      }
      return false;
    }
  }


  /**
   * Generates the queries to alter the tables and add the foreign keys and the description of the alterations
   *
   * @param {{string: [ForeignKey]}} constraintsToAdd
   * @returns {[[string], [string]]}
   */
  function generateQueries(constraintsToAdd) {
    const fkQueries = [];
    const fkDescription = [];

    Object.keys(constraintsToAdd).forEach(key => {
      const foreignKeys = constraintsToAdd[key];

      const fkAddString = [];

      foreignKeys.forEach(fk => fkAddString.push(getForeignKeyCreationString(fk)));

      fkQueries.push(`ALTER TABLE "${key}" ${fkAddString.join(', ')}`);
      fkDescription.push(`${key} (${foreignKeys.map(fk => fk.column)})`);
    });

    return [fkQueries, fkDescription];
  }


  /**
   * Runs the queries provided by generateQueries and displays the alterations
   *
   * @param {[string]} queries
   * @param {[string]} descriptions
   * @returns {Promise}
   */
  async function performQueries([queries, descriptions]) {

    if (queries.length < 1) {
      verbose && log(`No foreign keys to be added`);
      return undefined;
    }
    else {
      verbose && log(`Adding ${queries.length} foreign keys`);
    }

    try {
      await settings.connection.query(queries.join('; '));
      log(`Foreign Keys added: ${descriptions.join(', ')}`);
    }
    catch (e) {
      log(`Error while adding foreign keys. The structure for this table may be incorrect`);
      log(e.message);
    }
  }
}


/**
 * Searches for the indexes in the database and inserts them if necessary
 *
 * @returns {Promise}
 */
async function searchIndexes() {

  verbose && log(`Checking indexes and uniques`);

  const indexes = {};

  settings.structure.forEach(table => {
    let unique = [];
    let index  = [];

    if (table.unique && table.unique.length > 0) { unique = table.unique; }
    if (table.index  && table.index.length  > 0) { index  = table.index;  }

    if (unique.length > 0 || index.length > 0) { indexes[table.name] = { unique, index }; }
  });


  const statements = [];

  Object.keys(indexes).forEach(key => {
    const index = indexes[key].index.length > 0 ? (
      `(${indexes[key].index.map(key => "indexdef LIKE '%(" + key + ")%'").join(' OR ')})`
    ) : '';
    const unique = indexes[key].unique.length > 0 ? (
      `(${indexes[key].unique.map(key => "indexdef LIKE '%(" + key + ")%'").join(' OR ')})`
    ) : '';

    const or = indexes[key].index.length > 0 && indexes[key].unique.length > 0 ? ' OR ' : '';

    const statement = `SELECT * FROM pg_indexes WHERE tablename='${key}' AND (${unique} ${or} ${index});`;

    statements.push([statement, indexes[key], key]);
  });

  for (let i = 0; i < statements.length; i++) {
    try {
      await queryForIndex(statements[i]);
    }
    catch (e) {
      log(e);
    }
  }


  // ASYNC CALLS
  //--------------------------------------------------------------------------------------------------------------------

  async function queryForIndex([statement, table, tableName]) {
    try {
      const {rows} = await settings.connection.query(statement);
      const newIndexes = getNewConstraints(rows);
      const statements = getStatements(newIndexes);
      await runStatements(statements, newIndexes);
    }
    catch (e) {
      log(e);
    }


    /**
     * Receives the results form the searches and determines which constraints must be added
     *
     * @typedef {{Non_unique: number, Column_name: string}} row
     *
     * @param {[{
     *   indexdef: string,
     *   indexname: string,
     *   schemaname: string,
     *   tablename: string,
     * }]} rows
     * @returns {{unique: Array, index: Array}}
     */
    function getNewConstraints(rows) {
      const indexesInDb = [];
      const newIndexes = {
        unique: [],
        index:  []
      };

      rows.forEach(row => {
        const idx = /\(([^)]*)\)$/.exec(row.indexdef)[1];
        if (idx && !indexesInDb.includes(idx)) {
          indexesInDb.push(idx);
        }
      });

      newIndexes.unique = table.unique.filter(index => !indexesInDb.includes(index));
      newIndexes.index = table.index.filter(index => !indexesInDb.includes(index));

      return newIndexes;
    }


    /**
     * Gets the indexes generated by getNewConstraints and generates the statements, passing the indexes forward
     *
     * @param {{unique: Array, index: Array}} indexes
     * @returns {[string]}
     */
    function getStatements(indexes) {
      const statements = [];
      indexes.index.filter(i => !indexes.unique.includes(i)).forEach(i => statements.push(getIndexCreationString(i, 0, tableName)));
      indexes.unique.forEach(i => statements.push(getIndexCreationString(i, 1, tableName)));
      return statements;
    }


    /**
     * Runs the statements to add the constraints and displays the information messages
     *
     * @param {[string]} statements
     * @param {{unique: Array, index: Array}} indexes
     * @returns {Promise}
     */
    async function runStatements(statements, indexes) {
      if (statements.length < 1) {
        verbose && log(`No indexes to be added for table ${tableName}`);
        return undefined;
      }
      else {
        verbose && log(`Adding ${indexes.unique.length + indexes.index.length} indexes for table ${tableName}`);
      }

      const description = [];
      if (indexes.unique.length > 0) {
        description.push(`Uniques (${indexes.unique.join(', ')})`);
      }
      if (indexes.index.length > 0) {
        description.push(`Indexes (${indexes.index.join(', ')})`);
      }

      try {
        const query = statements.join('; ');
        await settings.connection.query(query);
        log(`Indexes added for table ${tableName}: ${description.join(', ')}`);
      }
      catch (e) {
        log(`Error while adding indexes and uniques. ` +
          `The structure for this table may be incorrect, or there are conflicting values in the database ` +
          `that blocking the creation of the constraint`);
        log(e.message);
      }
    }
  }
}


module.exports.connect = connect;
module.exports.expand = expand;