/**
 * Functions related to validation and generation of statements/queries
 *
 * @author Henrique Salvadori Coelho
 *         henriquesc@gmail.com
 *         http://hcoelho.com
 *         https://github.com/hscasn
 */

/**
 * Checks if a value is null or undefined
 *
 * @param {?} a value
 * @returns {boolean} true if it is null or undefined
 */
function isNullOrUndefined(a) {
  return a === undefined || a === null;
}


/**
 * Checks if a table is valid and should be searched
 *
 * @param {Table} table The structure of the table to be validated
 * @param {string?} table.name Name of the table
 * @param {boolean?} table.disabled Is the table disabled and should be ignored?
 * @returns {boolean} true if the table is valid and should be searched, false otherwise
 */
function isValidTable(table) {
  return (
      typeof table === 'object' &&
      !isNullOrUndefined(table.name) &&
      table.disabled !== true
  );
}


/**
 * Checks if a column is valid and should be searched/inserted
 *
 * @param {Column} column The structure of the column to be validated
 * @returns {boolean} true if the column is valid and should be searched, false otherwise
 */
function isValidColumn(column) {
  return (
      typeof column   === 'object' &&
      column.disabled !== true &&
      !isNullOrUndefined(column.type) &&
      !isNullOrUndefined(column.name)
  );
}


/**
 * Gets the query to crate a table
 *
 * @param {Table} table
 * @returns {string}
 */
function getCreationQuery(table) {
  if (!table || !table.columns || table.columns.constructor !== Array) { return ''; }

  const columns = table.columns.map(column => getColumnCreationString(column));

  return `CREATE TABLE "${table.name}" (${ columns.join(',') });`;
}


/**
 * Gets the query to alter a table and insert columns
 *
 * @param {Table} table
 * @param {[Column]} columns
 * @returns {string}
 */
function getAlterQuery(table, columns) {
  if (!table || !columns || columns.constructor !== Array) { return ''; }

  const sqlColumns = columns.map(column => ' ADD ' + getColumnCreationString(column));

  return `ALTER TABLE "${table.name}" ${ sqlColumns.join(',') };`;
}


/**
 * Generates the query for creating a column in a table
 *
 * @param {Column} column
 * @returns {string} query for column
 */
function getColumnCreationString(column) {

  if (!isValidColumn(column)) { return ''; }

  const size = column.size
      ? `(${column.size})`
      : '';

  const def = column.defaultValue !== undefined
    ? 'DEFAULT ' + column.defaultValue
    : '';

  const nullStr = column.isNull
      ? 'NULL'
      : 'NOT NULL';

  const primaryKey = column.primaryKey
      ? 'PRIMARY KEY'
      : '';

  return `"${column.name}" ${column.type}${size}  ${nullStr} ${def} ${primaryKey}`;
}


/**
 * Generates a statement to add a foreign key to a table
 *
 * @param {ForeignKey} foreignKey
 */
function getForeignKeyCreationString(foreignKey) {
  const onUpdate = getAction('ON UPDATE', foreignKey.onUpdate);
  const onDelete = getAction('ON DELETE', foreignKey.onDelete);

  return (
      `ADD CONSTRAINT ${foreignKey.column}_fk FOREIGN KEY (${foreignKey.column}) ` +
      `REFERENCES ${foreignKey.referenceTable}(${foreignKey.referenceColumn}) ` +
      `${onUpdate} ${onDelete}`
  );

  function getAction(prefix, action) {
    switch (action) {
      case 'setNull':     return prefix + ' SET NULL';
      case 'setDefault':  return prefix + ' SET DEFAULT';
      case 'restrict':    return prefix + ' RESTRICT';
      case 'cascade':     return prefix + ' CASCADE';
      case 'noAction':    return prefix + ' NO ACTION';
      default:            return '';
    }
  }
}


/**
 * Generates a statement to add an index/unique key to a table
 *
 * @param {string} constraint is the name of the column
 * @param {number} type specifies if the constraint is an index (0) or unique (1)
 * @param {string} table is the name of the table
 */
function getIndexCreationString(constraint, type, table) {
  return `CREATE ${getType(type)} INDEX ${constraint}_${table}_idx ON ${table} (${constraint})`;

  function getType(type) {
    switch (type) {
      case 0:  return '';
      case 1:  return 'UNIQUE';
      default: return '';
    }
  }
}


module.exports.isValidTable = isValidTable;
module.exports.getCreationQuery = getCreationQuery;
module.exports.getAlterQuery = getAlterQuery;
module.exports.getForeignKeyCreationString = getForeignKeyCreationString;
module.exports.getIndexCreationString = getIndexCreationString;